{
  Description:
    Tab set component that supports Left, Right, Top and Bottom tab orientations.
    Also provides the following properties:
      Notebook - Associated Notbbok control, for automatic page synch
      Selected/Unselected Color  of tabs
      TabRake and TabSpacing
      Accelerator characters supported on tabs
      Tab Focus supported
      Individual tab hints
      Automatic sizing of tab component to associated notebook

  Author: Thomas Hill, dba t.h.ink Software  (hilt@ione.net)
  History : Version 0.1 - 0.45
              Ancient history, best forgotten
            Version 0.5 - 01/11/96
              Total rewrite.
            Version : 0.52, 01/14/96
              Added Style property, for future OwnerDraw methods
              Added MainHintIsZero, to handle tab hints (ugly kludge!)
              Wrote help file for properties
              Added TabShadow property, for future 3D effects
              Rewrote tab dimensioning to use font size to compute tab sizes.
            Version 0.6, 1/16/96
              Surfaced OnChange property, identical to TTabSet operation (I hope)
              Added About box disguised as a component editor, adds an 'About...'
              item to the right-click menu.
}

unit Thtabs;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, StdCtrls, Dialogs, ExtCtrls;

type
  TOrientation = (toBottom,toRight,toTop,toLeft);
  TTabStyle = (tsStandard,tsOwnerDraw);
  TTabChangeEvent = procedure(Sender: TObject; NewTab: Integer; var AllowChange: Boolean) of object;
  TThTabSet = class(TCustomControl)
  private
{  Stuff that is published }
    FAutoSize : boolean;
    FBackColor : TColor;
    FChangeEvent : TTabChangeEvent;
    FCtrl3D : boolean;
    FMainHintisZero : boolean;
    FNoteBook : TNoteBook;
    FNoteColor : boolean;
    FOrientation : TOrientation;
    FSelectedColor : TColor;
    FStyle : TTabStyle;
    FTabList : TStrings;
    FTabHints : TStrings;
    FTabHeight : integer;
    FTabMargin : integer;
    FTabRake : integer;
    FTabShadow : boolean;
    FTabSpacing : integer;
    FTabIndex : integer;
    FTabFont : TFont;
    FUnSelectedColor : TColor;
{ stuff that is component internal }
    AllowChange : boolean;
    FTabRects : array[0..20] of TRect; { should make this a dynamic list, or something }

  protected
{ overridden parent procedures }
    procedure Click; override;
    procedure CreateParams(var Params : TCreateParams); override;
    procedure MouseMove(Shift : TShiftState; X, Y : integer); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Paint; override;
{ component specific procedures }
    procedure AdjustSize;
    procedure DrawTabs;
    procedure SetAutoSize(Value : boolean);
    procedure SetBackColor(Value : TColor);
    procedure SetCtrl3D(Value : boolean);
    procedure SetMainHintIsZero(Value : boolean);
    procedure SetNoteBook(NB : TNoteBook);
    procedure SetNoteColor(Value : boolean);
    procedure SetOrientation(Value : TOrientation);
    procedure SetSelColor(Value : TColor);
    procedure SetStyle(Value : TTabStyle);
    procedure SetTabSpacing(Value : integer);
    procedure SetTabHeight(Value : integer);
    procedure SetTabHints(List : TStrings);
    procedure SetTabList(List : TStrings);
    procedure SetTabIndex(Value : integer);
    procedure SetTabMargin(Value : integer);
    procedure SetTabRake(Value : integer);
    procedure SetTabShadow(Value : boolean);
    procedure SetUnSelColor(Value : TColor);
{ message handlers }
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure CMDialogChar(var Msg : TCMDialogChar); message cm_DialogChar;
  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;

    { Public declarations }
  published
    property AutoSize : boolean read FAutoSize write SetAutoSize default TRUE;
    property BackColor : TColor read FBackColor write SetBackColor default clGray;
    property Ctrl3D : boolean read FCtrl3D write SetCtrl3D default TRUE;
    property MainHintIsZero : boolean read FMainHintIsZero write SetMainHintIsZero default TRUE;
    property Notebook : TNoteBook read FNotebook write SetNotebook;
    property NotebookColor : boolean read FNoteColor write SetNoteColor default TRUE;
    property OnChange : TTabChangeEvent read FChangeEvent write FChangeEvent;
    property Orientation : TOrientation read FOrientation write SetOrientation default toBottom;
    property SelectedColor : TColor read FSelectedColor write SetSelColor default clTeal;
    property Spacing : integer read FTabSpacing write SetTabSpacing default 1;
    property Style : TTabStyle read FStyle write SetStyle default tsStandard;
    property Tabs : TStrings read FTabList write SetTabList;
    property TabHeight : integer read FTabHeight write SetTabHeight default 20;
    property TabHints : TStrings read FTabHints write SetTabHints;
    property TabIndex : integer read FTabIndex write SetTabIndex default 0;
    property TabMargin : integer read FTabMargin write SetTabMargin default 3;
    property TabRake : integer read FTabRake write SetTabRake default 5;
    property TabShadow : boolean read FTabShadow write SetTabShadow default FALSE;
    property UnSelectedColor : TColor read FUnSelectedColor write SetUnSelColor default clSilver;

  {inherited properties }
    property Align;
    property DragMode;
    property DragCursor;
    property Font;
    property Height default 30;
    property Hint;
    property ParentShowHint;
    property ShowHint;
    property TabOrder;
    property TabStop default TRUE;
    property Visible default TRUE;
    property Width default 200;
    property OnClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Custom', [TThTabSet]);
end;

constructor TThTabSet.Create(AOwner : TComponent);

begin
  inherited Create(AOwner);
  Width := 200;
  Height := 30;
  FOrientation := toBottom;
  FCtrl3D := TRUE;
  FTabShadow := FALSE;
  FTabMargin := 3;
  FTabRake := 5;
  FTabSpacing := 0;
  FTabHeight := 25;
  FTabIndex := 0;
  FStyle := tsStandard;
  TabStop := TRUE;
  FAutoSize := TRUE;
  FBackColor := clGray;
  FSelectedColor := clTeal;
  FUnSelectedColor := clSilver;
  FNotebook := nil;
  FTabList := TStringList.Create;
  FTabHints := TStringList.Create;
  FMainHintIsZero := TRUE;
  with Font do
  begin
    Name := 'MS Sans Serif';
    Height := -10;
    Color := clBlack;
    Style := [];
  end;
end;

destructor TThTabSet.Destroy;

var
   i : integer;

begin
  FTabList.Free;
  FTabHints.Free;
  if FNotebook <> nil then FNotebook := nil;
  inherited Destroy;
end;

procedure TThTabSet.CreateParams(var Params : TCreateParams);

begin
  inherited CreateParams(Params);
end;

procedure TThTabSet.SetOrientation(Value : TOrientation);

begin
  if FOrientation <> Value then
  begin
    FOrientation := Value;
    if FAutoSize and (FNotebook <> nil) then AdjustSize;
    Invalidate;
  end;
end;

procedure TThTabSet.SetBackColor(Value : TColor);

begin
  if FBackColor <> Value then
  begin
    FBackColor := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetTabHeight(Value : integer);

begin
  if FTabHeight <> Value then
  begin
    if Value < 0 then Value := 0;
    FTabHeight := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetTabMargin(Value : integer);

begin
  if FTabMargin <> Value then
  begin
    if Value < 0 then Value := 0;
    FTabMargin := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetTabRake(Value : integer);

begin
  if FTabRake <> Value then
  begin
    if Value < 0 then Value := 0;
    FTabRake := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetTabSpacing(Value : integer);

begin
  if FTabSpacing <> Value then
  begin
    FTabSpacing := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetStyle(Value : TTabStyle);

begin
  if FStyle <> Value then
  begin
    FStyle := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetSelColor(Value : TColor);

begin
  if FSelectedColor <> Value then
  begin
    FSelectedColor := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetAutoSize(Value : boolean);

begin
  if FAutoSize <> Value then
  begin
    FAutoSize := Value;
    AdjustSize;
    Invalidate;
  end;
end;

procedure TThTabSet.SetCtrl3D(Value : boolean);

begin
  if FCtrl3D <> Value then
  begin
    FCtrl3D := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetTabShadow(Value : boolean);

begin
  if FTabShadow <> Value then
  begin
    FTabShadow := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetUnSelColor(Value : TColor);

begin
  if FUnSelectedColor <> Value then
  begin
    FUnSelectedColor := Value;
    Invalidate;
  end;
end;

procedure TThTabSet.SetTabIndex(Value : integer);

begin
  if FTabIndex <> Value then
  begin
    AllowChange := TRUE;
    if assigned(FChangeEvent) then FChangeEvent(Self,Value,AllowChange);
    if AllowChange then
    begin
      if Value < 0 then Value := 0;
      FTabIndex := Value;
      Invalidate;
    end;
  end;
end;

procedure TThTabSet.SetTabList(List : TStrings);

begin
  if FTabList <> nil then
  begin
    FTabList.Clear;
  end;
  if List <> nil then
  begin
    FTabList.Assign(List);
  end;
  Invalidate;
end;

procedure TThTabSet.SetTabHints(List : TStrings);

begin
  if FTabHints <> nil then
  begin
    FTabHints.Clear;
    if List <> nil then FTabHints.Assign(List);
    Invalidate;
  end;
end;

procedure TThTabSet.SetMainHintIsZero(Value : boolean);

begin
  if FMainHintIsZero <> Value then
  begin
    FMainHintIsZero := Value;
  end;
end;

procedure TThTabSet.SetNoteBook(NB : TNoteBook);

begin
  if FNoteBook <> NB then
  begin
    FNoteBook := NB;
    FTabList.Clear;
    if NB <> nil then
    begin
      FTabList.Assign(NB.Pages);
      if FNoteColor then FSelectedColor := FNoteBook.Color;
      AdjustSize;
    end;
    Invalidate;
  end;
end;

procedure TThTabSet.SetNoteColor(Value : boolean);

begin
  if FNoteColor <> Value then
  begin
    FNoteColor := Value;
    if FNoteColor and (FNoteBook <> nil) then FSelectedColor := FNoteBook.Color;
    Invalidate;
  end;
end;

procedure TThTabSet.MouseMove(Shift : TShiftState; X, Y : integer);

var
   Found : boolean;
   i : integer;

begin
  inherited MouseMove(Shift,X,Y);
  Hint := ''; Found := FALSE;
  if (FTabList.Count > 0) and (FTabHints.Count > 0) and ShowHint then
  begin
    for i := 0 to FTabHints.Count - 1 do
    begin
      if PtInRect(FTabRects[i],Point(X,Y)) then
      begin
        if (FTabHints.Count > 0) and (length(FTabHints[i]) > 0) then
        begin
          if FMainHintIsZero then
            Hint := FTabHints[i + 1]
          else
            Hint := FTabHints[i];
          Found := TRUE;
          break;
        end;
      end;
    end;
    if not Found and FMainHintIsZero then Hint := FTabHints[0];
  end;
end;


procedure TThTabSet.WMSetFocus(var Message: TWMSetFocus);

begin
  Invalidate;
end;

procedure TThTabSet.WMKillFocus(var Message: TWMKillFocus);

begin
  Invalidate;
end;

{ this is important in order to safely remove the reference to the associated
  notebook component if it disappears }
procedure TThTabSet.Notification(AComponent: TComponent; Operation: TOperation);

begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FNotebook) and (Operation = opRemove) then FNotebook := nil
end;

{ This provides entry to accel chars from the main form keyboard loop }
procedure TThTabSet.CMDialogChar( var Msg : TCMDialogChar );

var
   i : integer;

begin
  with Msg do
  begin
    if FTabList.Count > 0 then
    begin
      for i := 0 to FTabList.Count - 1 do
      begin
       if IsAccel(CharCode,FTabList[i]) then
       begin
         if assigned(FChangeEvent) then FChangeEvent(Self,i,AllowChange);
         if AllowChange then
         begin
           FTabIndex := i;
           SetFocus;
           Invalidate;
         end;
         Result := 1;  { say we handled it, even if AllowChange is false }
         break;
       end;
      end;
    end
    else
      inherited;
  end;
end;

procedure TThTabSet.Click;

var
   MP,TP : TPoint;
   i : integer;

begin
  inherited Click;
  GetCursorPos(MP);
  TP := ScreenToClient(MP);
  if FTabList.Count > 0 then
  begin
    for i := 0 to FTabList.Count - 1 do
    begin
      if PtInRect(FTabRects[i],TP) then
      begin
        SetTabIndex(i);
        SetFocus;
      end;
    end;
    Invalidate;
  end;
end;

procedure TThTabSet.AdjustSize;

begin
  if FAutoSize and (FNotebook <> nil) then
  begin
    case FOrientation of
      toBottom : begin
                   Left := FNoteBook.Left;
                   Top := FNoteBook.Top + FNoteBook.Height;
                   Width := FNoteBook.Width;
                   Height := MulDiv(FTabHeight,150,100);
                 end;
        toLeft : begin
                   Left := FNoteBook.Left - MulDiv(Canvas.TextWidth('W'),250,100);
                   Top := FNotebook.Top;
                   Canvas.Font := Self.Font;
                   Width := MulDiv(Canvas.TextWidth('W'),300,100);
                   Height := FNotebook.Height;
                 end;
       toRight : begin
                   Left := FNotebook.Left + FNotebook.Width;
                   Top := FNoteBook.Top;
                   canvas.Font := Self.Font;
                   Width := MulDiv(Canvas.TextWidth('W'),300,100);
                   Height := FNotebook.Height;
                end;
        toTop : begin
                  Top := FNoteBook.Top - MulDiv(FTabHeight,150,100);
                  Left := FNoteBook.Left;
                  Width := FNoteBook.Width;
                  Height := MulDiv(FTabHeight,150,100);
                end;
    end; { Orientation case }
  end;
end;

procedure TThTabSet.Paint;

begin
  Drawtabs;
end;

procedure TThTabSet.DrawTabs;

var
   CharW,CharH : integer;
   R : TRect;
   S : string;
   TabCnt,CurTabH,CurTabW : integer;
   Cx,Cy,i,CharS : integer;
   Start,Corner1,Corner2,
   Corner3,Corner4,EndPnt : TPoint;
   FocusPt1,FocusPt2 : TPoint;
   pStr : array[0..2] of char;
   L : TLabel;

begin
  R := GetClientRect;
  Frame3D(Canvas,R,clBlack,clWhite,1);
  with Canvas do
  begin
    Font := Self.Font;
    Brush.Color := FBackColor;
    FillRect(R);
    CharW := TextWidth('W');
    CharH := TextHeight('W');
    CurTabH := CharH * 2;
    CurTabW := CharW * 2;
    if FTabList.Count > 0 then
    begin
      case FOrientation of
        toBottom : begin
                     R := GetClientRect;
                     Start := Point(0,0);
                     EndPnt := Point(R.Right,0);
                     Cx := FTabMargin; Cy := CurTabH;
                     Pen.Color := clBlack;
                     for TabCnt := 0 to FTabList.Count - 1 do
                     begin
                       S := FTabList[TabCnt];
                       Corner1 := Point(Cx,0);
                       Corner2 := Point(Cx + FTabRake,CurTabH);
                       if FCtrl3D then inc(Corner2.X,3);
                       Corner3 := Point(Corner2.X + MulDiv(TextWidth(S),150,100),Cy);
                       if (length(S) = 1) or ((S[1] = '&') and (length(S) = 2)) then
                         inc(Corner3.X,CharW);
                       Corner4 := Point(Corner3.X + FTabRake,0);
                       if FCtrl3D then inc(Corner4.X,3);
                       Pen.Color := clBlack; Pen.Width := 1;
                       PolyLine([Start,Corner1,Corner2,Corner3,Corner4,EndPnt]);
                       if TabCnt <> FTabIndex then
                         Brush.Color := FUnSelectedColor
                       else
                         Brush.Color := FSelectedColor;
                       FloodFill(Corner2.X + 2,Corner2.Y - 2,clBlack,fsBorder);
                       if Ctrl3D then
                       begin
                         Pen.Color := clWhite;
                         Pen.Width := 2;
                         inc(Corner1.X);
                         inc(Corner2.X); dec(Corner2.Y);
                         PolyLine([Corner1,Corner2]);
                         Pen.Color := clGray;
                         Dec(Corner3.X); dec(Corner3.Y);
                         dec(Corner4.X);
                         PolyLine([Corner2,Corner3,Corner4]);
                         Pen.Width := 1;
                       end;
                       if TabCnt = FTabIndex then
                       begin
                         FocusPt1 := Corner1;
                         FocusPt2 := Corner4;
                       end;
                       R := Rect(Corner2.X,Corner1.Y,Corner3.X,Corner3.Y);
                       InflateRect(R,-2,-2);
                       inc(R.Left); inc(R.Top);
                       DrawText(Handle,@S[1],length(S),R,
                                DT_VCENTER or DT_CENTER or DT_SINGLELINE);
                       FTabRects[TabCnt] := R;
                       inc(Cx,((Corner4.X - Corner1.X) + FTabSpacing));
                     end;
                     if FTabIndex >= 0 then
                     begin
                       MoveTo(FocusPt1.X,FocusPt1.Y);
                       Pen.Color := FSelectedColor;
                       LineTo(FocusPt2.X,FocusPt2.Y);
                     end;
                   end;
           toTop : begin
                     R := GetClientRect;
                     Start := Point(0,Height - 1);
                     EndPnt := Point(R.Right,Height - 1);
                     Cx := FTabMargin; Cy := Height - CurTabH - 1;
                     Pen.Color := clBlack; Pen.Width := 1;
                     for TabCnt := 0 to FTabList.Count - 1 do
                     begin
                       S := FTabList[TabCnt];
                       Corner1 := Point(Cx,Height - 1);
                       Corner2 := Point(Cx + FTabRake,Cy);
                       if FCtrl3D then inc(Corner2.X,3);
                       Corner3 := Point(Corner2.X + MulDiv(TextWidth(S),150,100),Cy);
                       if (length(S) = 1) or ((S[1] = '&') and (length(S) = 2)) then
                         inc(Corner3.X,CharW);
                       Corner4 := Point(Corner3.X + FTabRake,Height - 1);
                       if FCtrl3D then inc(Corner4.X,3);
                       Pen.Color := clWhite; Pen.Width := 1;
                       PolyLine([Start,Corner1]);
                       Pen.Color := clBlack;
                       PolyLine([Corner1,Corner2,Corner3,Corner4]);
                       Pen.Color := clWhite;
                       PolyLine([Corner4,EndPnt]);
                       if TabCnt <> FTabIndex then
                         Brush.Color := FUnSelectedColor
                       else
                         Brush.Color := FSelectedColor;
                       FloodFill(Corner2.X + 2,Corner2.Y + 2,FBackColor,fsSurface);
                       if Ctrl3D then
                       begin
                         Pen.Color := clWhite;
                         Pen.Width := 2;
                         inc(Corner1.X,2);
                         inc(Corner2.X,2); inc(Corner2.Y,2);
                         Dec(Corner3.X); inc(Corner3.Y);
                         PolyLine([Corner1,Corner2,Corner3]);
                         Pen.Color := clGray;
                         dec(Corner4.X);
                         PolyLine([Corner3,Corner4]);
                         Pen.Width := 1;
                       end;
                       if TabCnt = FTabIndex then
                       begin
                         FocusPt1 := Corner1;
                         FocusPt2 := Corner4;
                       end;
                       R := Rect(Corner2.X,Corner2.Y,Corner3.X,Corner4.Y);
                       InflateRect(R,-2,-2);
                       inc(R.Left); inc(R.Top);
                       DrawText(Handle,@S[1],length(S),R,
                                DT_VCENTER or DT_CENTER or DT_SINGLELINE);
                       FTabRects[TabCnt] := R;
                       inc(Cx,((Corner4.X - Corner1.X) + FTabSpacing));
                     end;
                     if FTabIndex >= 0 then
                     begin
                       MoveTo(FocusPt1.X,FocusPt1.Y);
                       Pen.Color := FSelectedColor;
                       LineTo(FocusPt2.X,FocusPt2.Y);
                     end;
                   end;
         toRight : begin
                     R := GetClientRect;
                     Start := Point(0,0);
                     EndPnt := Point(0,R.Bottom);
                     Cx := CurTabW; Cy := FTabMargin;
                     Pen.Color := clBlack;
                     for TabCnt := 0 to FTabList.Count - 1 do
                     begin
                       S := FTabList[TabCnt];
                       Corner1 := Point(0,Cy);
                       Corner2 := Point(Cx,Cy + FTabRake);
                       if FCtrl3D then inc(Corner2.Y,3);
                       Corner3 := Point(Corner2.X ,Corner2.Y + (CharH * length(S)) + CharH);
                       for i := 1 to length(S) do
                       begin
                         if S[i] = '&' then
                         begin
                           dec(Corner3.Y,CharH); break;
                         end;
                       end;
                       Corner4 := Point(0,Corner3.Y + FTabRake);
                       if FCtrl3D then inc(Corner4.Y,3);
                       Pen.Color := clBlack; pen.Width := 1;
                       PolyLine([Start,Corner1,Corner2,Corner3,Corner4,EndPnt]);
                       if TabCnt <> FTabIndex then
                         Brush.Color := FUnSelectedColor
                       else
                         Brush.Color := FSelectedColor;
                       FloodFill(Corner2.X - 2,Corner2.Y + 2,FBackColor,fsSurface);
                       if TabCnt = FTabIndex then
                       begin
                         FocusPt1 := Corner1;
                         FocusPt2 := Corner4;
                       end;
                       if Ctrl3D then
                       begin
                         Pen.Color := clWhite; Pen.Width := 2;
                         inc(Corner1.Y,2);
                         inc(Corner2.Y,2); dec(Corner2.X);
                         PolyLine([Corner1,Corner2]);
                         Pen.Color := clGray;
                         Dec(Corner3.X); dec(Corner3.Y);
                         dec(Corner4.Y,2);
                         PolyLine([Corner2,Corner3,Corner4]);
                         Pen.Width := 1;
                       end;
                       i := 1; CharS := Corner2.Y + CharH div 2;
                       repeat
                         R := Rect(0,CharS,Corner2.X,CharS + TextHeight(S[i]));
                         if S[i] = '&' then
                           StrPCopy(pStr,Copy(S,i,2))
                         else
                           StrPCopy(pStr,S[i]);
                         DrawText(Handle,pStr,strlen(pStr),R,DT_CENTER or DT_VCENTER or DT_SINGLELINE);
                         inc(CharS, CharH);
                         inc(i,strlen(pStr));
                       until i > length(S);
                       R := Rect(0,Corner2.Y,Corner3.X,Corner3.Y);
                       InflateRect(R,-2,0);
                       inc(R.Left);
                       FTabRects[TabCnt] := R;
                       inc(Cy,((Corner4.Y - Corner1.Y) + FTabSpacing));
                     end;
                     if FTabIndex >= 0 then
                     begin
                       MoveTo(FocusPt1.X,FocusPt1.Y);
                       Pen.Color := FSelectedColor;
                       LineTo(FocusPt2.X,FocusPt2.Y);
                     end;
                   end;
          toLeft : begin
                     R := GetClientRect;
                     Start := Point(Width - 1,0);
                     EndPnt := Point(Width - 1,R.Bottom);
                     Cx := Width - CurTabW; Cy := FTabMargin;
                     for TabCnt := 0 to FTabList.Count - 1 do
                     begin
                       S := FTabList[TabCnt];
                       Corner1 := Point(Width - 1,Cy);
                       Corner2 := Point(Cx,Cy + FTabRake);
                       if FCtrl3D then inc(Corner2.Y,3);
                       Corner3 := Point(Corner2.X ,Corner2.Y + (CharH * length(S)) + CharH);
                       for i := 1 to length(S) do
                       begin
                         if S[i] = '&' then
                         begin
                           dec(Corner3.Y,CharH); break;
                         end;
                       end;
                       Corner4 := Point(Width - 1,Corner3.Y + FTabRake);
                       if FCtrl3D then inc(Corner4.Y,3);
                       Pen.Color := clWhite; Pen.Width := 1;
                       PolyLine([Start,Corner1]);
                       Pen.Color := clBlack;
                       PolyLine([Corner1,Corner2,Corner3,Corner4]);
                       Pen.Color := clWhite;
                       PolyLine([Corner4,EndPnt]);
                       if TabCnt <> FTabIndex then
                         Brush.Color := FUnSelectedColor
                       else
                         Brush.Color := FSelectedColor;
                       FloodFill(Corner2.X + 2,Corner2.Y + 2,FBackColor,fsSurface);
                       if TabCnt = FTabIndex then
                       begin
                         FocusPt1 := Corner1;
                         FocusPt2 := Corner4;
                       end;
                       if Ctrl3D then
                       begin
                         Pen.Color := clWhite; Pen.Width := 2;
                         inc(Corner1.Y,2);
                         inc(Corner2.Y,2); inc(Corner2.X,2);
                         dec(Corner3.Y); inc(Corner3.X,2);
                         PolyLine([Corner1,Corner2,Corner3]);
                         Pen.Color := clGray;
                         dec(Corner4.Y,2); dec(corner4.X);
                         PolyLine([Corner3,Corner4]);
                         Pen.Width := 1;
                       end;
                       i := 1; CharS := Corner2.Y + CharH div 2;
                       repeat
                         R := Rect(Corner2.X,CharS,Width,CharS + TextHeight(S[i]));
                         if S[i] = '&' then
                           StrPCopy(pStr,Copy(S,i,2))
                         else
                           StrPCopy(pStr,S[i]);
                         DrawText(Handle,pStr,strlen(pStr),R,DT_CENTER or DT_VCENTER or DT_SINGLELINE);
                         inc(CharS, CharH);
                         inc(i,strlen(pStr));
                       until i > length(S);
                       R := Rect(Cx,Corner2.Y,Corner4.X,Corner3.Y);
                       InflateRect(R,-2,0);
                       dec(R.Right);
                       FTabRects[TabCnt] := R;
                       inc(Cy,((Corner4.Y - Corner1.Y) + FTabSpacing));
                     end;
                     if FTabIndex >= 0 then
                     begin
                       MoveTo(FocusPt1.X,FocusPt1.Y);
                       Pen.Color := FSelectedColor;
                       LineTo(FocusPt2.X,FocusPt2.Y);
                     end;
                   end;
      end; { orientation case }
      if Focused then
      begin
        R := FTabRects[FTabIndex];
        DrawFocusRect(R);
      end;
      if Assigned(FNoteBook) then FNotebook.PageIndex := FTabIndex;
    end; { have tabs to show }
  end; { canvas containment }
end;

end.
